"""Classes related to card endpoints
"""
from __future__ import annotations

from datetime import datetime
from typing import Any, ClassVar, Optional
from uuid import UUID

from pydantic.fields import Field

from metabase_tools.exceptions import (
    EmptyDataReceived,
    InvalidDataReceived,
    RequestFailure,
)
from metabase_tools.metabase import MetabaseApi
from metabase_tools.models.collection import Collection
from metabase_tools.models.generic import GA, GenericWithArchive
from metabase_tools.models.user import User


class Card(GenericWithArchive):
    """Card object class with related methods"""

    BASE_EP: ClassVar[str] = "/card"

    description: Optional[str]
    archived: bool
    collection_position: Optional[int]
    table_id: Optional[int]
    result_metadata: Optional[list[dict[str, Any]]]
    creator: User
    database_id: Optional[int]
    enable_embedding: bool
    collection_id: Optional[int]
    query_type: Optional[str]
    creator_id: int
    updated_at: datetime
    made_public_by_id: Optional[int]
    embedding_params: Optional[dict[str, Any]]
    cache_ttl: Optional[str]
    dataset_query: dict[str, Any]
    display: str
    last_edit_info: Optional[dict[str, Any]] = Field(alias="last-edit-info")
    visualization_settings: dict[str, Any]
    collection: Optional[Collection]
    dataset: Optional[int]
    created_at: datetime
    public_uuid: Optional[UUID]
    can_write: Optional[bool]
    dashboard_count: Optional[int]
    is_favorite: Optional[bool] = Field(alias="favorite")

    @classmethod
    def related(
        cls: type[GA], adapter: MetabaseApi, targets: list[int]
    ) -> list[dict[str, Any]]:
        """Objects related to targets

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): Card IDs to pull

        Returns:
            list[dict]: List of dicts with related objects for each target
        """
        results = []
        for target in targets:
            new = {"card_id": target}
            result = adapter.get(endpoint=f"/card/{target}/related")
            if isinstance(result, dict):
                new |= result
            results.append(new)
        return results

    @classmethod
    def embeddable(cls: type[GA], adapter: MetabaseApi) -> list[GA]:
        """Fetch list of cards with embedding enabled

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Raises:
            EmptyDataReceived: If no cards have embedding enabled

        Returns:
            list[Self]: List of cards with embedding enabled
        """
        cards = adapter.get(endpoint="/card/embeddable")
        if cards:
            return [cls(**card) for card in cards if isinstance(card, dict)]
        raise EmptyDataReceived

    @classmethod
    def favorite(
        cls: type[GA], adapter: MetabaseApi, targets: list[int]
    ) -> list[dict[str, Any]]:
        """Favorite cards

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): Card IDs to favorite

        Returns:
            list[dict]: Results of favoriting operation
        """
        results = []
        for target in targets:
            try:
                result = adapter.post(endpoint=f"/card/{target}/favorite")
            except RequestFailure:
                result = {
                    "card_id": target,
                    "error": "Metabase error, probably already a favorite",
                }
            if isinstance(result, dict):
                results.append(result)
        return results

    @classmethod
    def unfavorite(
        cls: type[GA], adapter: MetabaseApi, targets: list[int]
    ) -> list[dict[str, Any]]:
        """Unfavorite cards

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): Card IDs to unfavorite

        Returns:
            list[dict]: Results of unfavoriting operation
        """
        results = []
        for target in targets:
            result: dict[str, int | bool | str] = {}
            try:
                _ = adapter.delete(endpoint=f"/card/{target}/favorite")
                result = {"card_id": target, "success": True}
            except (InvalidDataReceived, RequestFailure):
                result = {
                    "card_id": target,
                    "error": "Metabase error, probably not a favorite",
                }
            results.append(result)
        return results

    @classmethod
    def share(
        cls: type[GA], adapter: MetabaseApi, targets: list[int]
    ) -> list[dict[str, Any]]:
        """Generate publicly-accessible links for cards

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): Card IDs to share

        Returns:
            list[dict]: UUIDs to be used in public links.
        """
        return cls._request_list(
            http_method="POST",
            adapter=adapter,
            endpoint="/card/{id}/public_link",
            source=targets,
        )

    @classmethod
    def unshare(
        cls: type[GA], adapter: MetabaseApi, targets: list[int]
    ) -> list[dict[str, Any]]:
        """Remove publicly-accessible links for cards

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): Card IDs to unshare

        Returns:
            list[dict]: Result of unshare operation
        """
        return cls._request_list(
            http_method="DELETE",
            adapter=adapter,
            endpoint="/card/{id}/public_link",
            source=targets,
        )



"""Classes related to collections endpoints
"""
from __future__ import annotations  # Included for support of |

from typing import Any, ClassVar, Optional

from metabase_tools.exceptions import EmptyDataReceived
from metabase_tools.metabase import MetabaseApi
from metabase_tools.models.generic import GA, GenericWithArchive


class Collection(GenericWithArchive):
    """Collection object class with related methods"""

    BASE_EP: ClassVar[str] = "/collection"

    description: Optional[str]
    archived: Optional[bool]
    slug: Optional[str]
    color: Optional[str]
    personal_owner_id: Optional[int]
    location: Optional[str]
    namespace: Optional[int]
    effective_location: Optional[str]
    effective_ancestors: Optional[list[dict[str, Any]]]
    can_write: Optional[bool]
    parent_id: Optional[int]

    @classmethod
    def get_tree(cls: type[GA], adapter: MetabaseApi) -> list[dict[str, Any]]:
        """Collection tree

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Raises:
            EmptyDataReceived: No data returned from API

        Returns:
            list[dict]: Representation of collection tree
        """
        response = adapter.get(endpoint="/collection/tree")
        if isinstance(response, list) and all(
            isinstance(record, dict) for record in response
        ):
            return response
        raise EmptyDataReceived

    @staticmethod
    def _flatten_tree(parent: dict[str, Any], path: str = "/") -> list[dict[str, Any]]:
        """Recursive function to flatten collection tree to show the full path for all\
             collections

        Args:
            parent (dict): Parent collection
            path (str, optional): Path to parent collection. Defaults to "/".

        Returns:
            list[dict]: Flattened list
        """
        children = []
        for child in parent["children"]:
            children.append(
                {
                    "id": child["id"],
                    "name": child["name"],
                    "path": f'{path}/{parent["name"]}/{child["name"]}'.replace(
                        "//", "/"
                    ),
                }
            )
            if "children" in child and len(child["children"]) > 0:
                grandchildren = Collection._flatten_tree(
                    child, f'{path}/{parent["name"]}'.replace("//", "/")
                )
                if isinstance(grandchildren, list):
                    children.extend(grandchildren)
                else:
                    children.append(grandchildren)
        return children

    @classmethod
    def get_flat_list(cls, adapter: MetabaseApi) -> list[dict[str, Any]]:
        """Flattens collection tree so the full path of each collection is shown

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Returns:
            list[dict]: Flattened collection tree
        """
        tree = cls.get_tree(adapter=adapter)
        folders = []
        for root_folder in tree:
            if root_folder["personal_owner_id"] is not None:  # Skips personal folders
                continue
            folders.append(
                {
                    "id": root_folder["id"],
                    "name": root_folder["name"],
                    "path": f'/{root_folder["name"]}',
                }
            )
            folders.extend(Collection._flatten_tree(root_folder))
        return folders

    @classmethod
    def get_contents(
        cls,
        adapter: MetabaseApi,
        collection_id: int,
        model_type: Optional[str] = None,
        archived: bool = False,
    ) -> list[dict[str, Any]]:
        """Get the contents of the provided collection

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            collection_id (int): ID of the requested collection
            model_type (str, optional): Filter to provided model. Defaults to all.
            archived (bool, optional): Archived objects. Defaults to False.

        Raises:
            EmptyDataReceived: No results from API

        Returns:
            list: Contents of collection
        """
        params: dict[Any, Any] = {}
        if archived:
            params["archived"] = archived
        if model_type:
            params["model"] = model_type

        response = adapter.get(
            endpoint=f"/collection/{collection_id}/items",
            params=params,
        )

        if isinstance(response, list) and all(
            isinstance(record, dict) for record in response
        ):
            return response
        raise EmptyDataReceived

    @classmethod
    def graph(cls, adapter: MetabaseApi) -> dict[str, Any]:
        """Graph of collection

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Returns:
            dict: graph of collection
        """
        result = adapter.get(endpoint="/collection/graph")
        if isinstance(result, dict):
            return result
        raise EmptyDataReceived



"""Unofficial API wrapper for Metabase
"""

__version__ = "0.6.0"

from metabase_tools.exceptions import (
    AuthenticationFailure,
    EmptyDataReceived,
    InvalidDataReceived,
    InvalidParameters,
    ItemInPersonalCollection,
    ItemNotFound,
    MetabaseApiException,
    NoUpdateProvided,
    RequestFailure,
)
from metabase_tools.metabase import MetabaseApi
from metabase_tools.models.card import Card
from metabase_tools.models.collection import Collection
from metabase_tools.models.database import Database
from metabase_tools.models.user import User
from metabase_tools.tools import MetabaseTools

__all__ = (
    "AuthenticationFailure",
    "EmptyDataReceived",
    "InvalidDataReceived",
    "InvalidParameters",
    "ItemNotFound",
    "ItemInPersonalCollection",
    "MetabaseApiException",
    "NoUpdateProvided",
    "RequestFailure",
    "MetabaseApi",
    "Card",
    "Collection",
    "Database",
    "User",
    "MetabaseTools",
)



"""Classes related to database endpoints
"""

from datetime import datetime
from typing import Any, ClassVar, Optional

from metabase_tools.models.generic import GenericWithoutArchive


class Database(GenericWithoutArchive):
    """Database object class with related methods"""

    BASE_EP: ClassVar[str] = "/database"

    description: Optional[str]
    features: list[str]
    cache_field_values_schedule: str
    timezone: str
    auto_run_queries: bool
    metadata_sync_schedule: str
    caveats: Optional[str]
    is_full_sync: bool
    updated_at: datetime
    native_permissions: Optional[str]
    details: dict[str, Any]
    is_sample: bool
    is_on_demand: bool
    options: Optional[str]
    engine: str
    refingerprint: Optional[str]
    created_at: datetime
    points_of_interest: Optional[str]
    schedules: Optional[dict[str, Any]]



"""Classes related to user endpoints
"""

from datetime import datetime
from typing import Any, ClassVar, Optional

from pydantic.fields import Field

from metabase_tools.exceptions import RequestFailure
from metabase_tools.metabase import MetabaseApi
from metabase_tools.models.generic import GWA, GenericWithoutArchive


class User(GenericWithoutArchive):
    """User object class with related methods"""

    BASE_EP: ClassVar[str] = "/user"

    name: str = Field(alias="common_name")
    email: str
    first_name: str
    last_name: str
    date_joined: datetime
    last_login: Optional[datetime]
    updated_at: Optional[datetime]
    is_qbnewb: bool
    is_superuser: bool
    ldap_auth: Optional[bool]
    google_auth: Optional[bool]
    is_active: Optional[bool]
    locale: Optional[str]
    group_ids: Optional[list[int]]
    login_attributes: Optional[list[dict[str, Any]]]
    personal_collection_id: Optional[int]

    @classmethod
    def current(cls: type[GWA], adapter: MetabaseApi) -> GWA:
        """Current user details

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Raises:
            RequestFailure: Invalid response from API

        Returns:
            User: Current user details
        """
        response = adapter.get(endpoint="/user/current")
        if isinstance(response, dict):
            return cls(**response)
        raise RequestFailure

    @classmethod
    def delete(cls, adapter: MetabaseApi, targets: list[int]) -> dict[int, Any]:
        """Disables user(s) provided

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of users to disable

        Returns:
            dict: Dict of users that were disabled with results
        """
        # This method is provided to override the super class' delete function
        return cls.disable(adapter=adapter, targets=targets)

    @classmethod
    def disable(cls, adapter: MetabaseApi, targets: list[int]) -> dict[int, Any]:
        """Disables user(s) provided

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of users to disable

        Returns:
            dict: Dict of users that were disabled with results
        """
        return super(User, cls).delete(adapter=adapter, targets=targets)

    @classmethod
    def enable(cls: type[GWA], adapter: MetabaseApi, targets: list[int]) -> list[GWA]:
        """Enable user(s) provided

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of users to enable

        Returns:
            list[User]: Enabled users
        """
        results = cls._request_list(
            http_method="PUT",
            adapter=adapter,
            endpoint="/user/{id}/reactivate",
            source=[{"id": target} for target in targets],
        )
        return [cls(**result) for result in results]

    @classmethod
    def resend_invite(
        cls: type[GWA], adapter: MetabaseApi, targets: list[int]
    ) -> list[GWA]:
        """Resent user invites

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of users to resend invites for

        Returns:
            list[User]: Users with a resent invite
        """
        results = cls._request_list(
            http_method="PUT",
            adapter=adapter,
            endpoint="/user/{id}/send_invite",
            source=[{"id": target} for target in targets],
        )
        return [cls(**result) for result in results]

    @classmethod
    def update_password(
        cls: type[GWA], adapter: MetabaseApi, payloads: list[dict[str, Any]]
    ) -> list[GWA]:
        """Updates passwords for users

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            payloads (list[dict]): List of users and passwords to update

        Returns:
            list[User]: List of users with reset passwords
        """
        results = cls._request_list(
            http_method="PUT",
            adapter=adapter,
            endpoint="/user/{id}/password",
            source=payloads,
        )
        return [cls(**result) for result in results]

    @classmethod
    def qbnewb(cls: type[GWA], adapter: MetabaseApi, targets: list[int]) -> list[GWA]:
        """Indicate that a user has been informed about Query Builder.

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of users to toggle

        Returns:
            list[Self]: Users with query builder toggle set
        """
        results = cls._request_list(
            http_method="PUT",
            adapter=adapter,
            endpoint="/user/{id}/qbnewb",
            source=[{"id": target} for target in targets],
        )
        return [cls(**result) for result in results]



"""
Rest adapter for the Metabase API
"""
from __future__ import annotations  # Included for support of |

import logging
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Optional

from requests import Response, Session
from requests.exceptions import RequestException

from metabase_tools.exceptions import (
    AuthenticationFailure,
    InvalidDataReceived,
    RequestFailure,
)


class MetabaseApi:
    """Metabase API adapter"""

    def __init__(
        self,
        metabase_url: str,
        credentials: Optional[dict[str, str]] = None,
        cache_token: bool = False,
        token_path: Optional[Path | str] = None,
    ):
        self._logger = logging.getLogger(__name__)
        if not credentials and not token_path:
            raise AuthenticationFailure("No authentication method provided")
        credentials = credentials or {}
        token_path = Path(token_path) if token_path else None

        # Validate Metabase URL
        self.metabase_url = self._validate_base_url(url=metabase_url)

        # Starts session to be reused by the adapter so that the auth token is cached
        self._session = Session()

        # Authenticate
        self._authenticate(token_path=token_path, credentials=credentials)

        if cache_token:
            save_path = Path(token_path or "metabase.token")
            self.save_token(save_path=save_path)

    def _validate_base_url(self, url: str) -> str:
        if url[-1] == "/":
            url = url[:-1]
        if url[-4:] == "/api":
            url = url[:-4]
        if url[:4] != "http":
            url = f"http://{url}"
        return f"{url}/api"

    def _authenticate(
        self, token_path: Optional[Path], credentials: dict[str, str]
    ) -> None:
        authed = False
        # Try cached token first
        if token_path and token_path.exists():
            authed = self._auth_with_cached_token(token_path=token_path)
            if not authed:
                self._delete_cached_token(token_path=token_path)
        # Try token passed as credentials next
        if not authed and "token" in credentials:
            authed = self._auth_with_passed_token(credentials=credentials)
        # Finally try username and password
        if not authed and "username" in credentials and "password" in credentials:
            authed = self._auth_with_login(credentials=credentials)
        # Raise error if still not authenticated
        if not authed:
            self._logger.error("Failed to authenticate")
            raise AuthenticationFailure(
                "Failed to authenticate with credentials provided"
            )

    def _add_token_to_header(self, token: str) -> None:
        headers = {
            "Content-Type": "application/json",
            "X-Metabase-Session": token,
        }
        self._session.headers.update(headers)

    def _delete_cached_token(self, token_path: Path) -> None:
        if token_path.exists():
            self._logger.warning("Deleting token file")
            token_path.unlink()

    def _auth_with_cached_token(self, token_path: Path) -> bool:
        with open(token_path, "r", encoding="utf-8") as file:
            token = file.read()
        self._logger.debug("Attempting authentication with token file")
        self._add_token_to_header(token=token)
        authed = self.test_for_auth()
        self._logger.debug(
            "Authenticated with token file"
            if authed
            else "Failed to authenticate with token file"
        )
        return authed

    def _auth_with_passed_token(self, credentials: dict[str, str]) -> bool:
        self._logger.debug("Attempting authentication with token passed")
        self._add_token_to_header(token=credentials["token"])
        authed = self.test_for_auth()
        self._logger.debug(
            "Authenticated with token passed"
            if authed
            else "Failed to authenticate with token passed"
        )
        return authed

    def _auth_with_login(self, credentials: dict[str, str]) -> bool:
        """Private method for authenticating a session with the API

        Args:
            credentials (dict): Username and password
        """
        try:
            self._logger.debug("Attempting authentication with username and password")
            response = self._session.post(
                f"{self.metabase_url}/session", json=credentials
            )
            self._add_token_to_header(token=response.json()["id"])
            authed = self.test_for_auth()
            self._logger.debug(
                "Authenticated with login"
                if authed
                else "Failed to authenticate with login"
            )
            return authed
        except KeyError as error_raised:
            self._logger.warning(
                "Exception encountered during attempt to authenticate with login \
                    passed: %s",
                error_raised,
            )
        return False

    def test_for_auth(self) -> bool:
        """Validates successful authentication by attempting to retrieve data about \
            the current user

        Returns:
            bool: Successful authentication
        """
        return (
            200
            <= self._session.get(f"{self.metabase_url}/user/current").status_code
            <= 299
        )

    def save_token(self, save_path: Path | str) -> None:
        """Writes active token to the specified file

        Args:
            save_path (Path | str): Name of file to write
        """
        token = str(self._session.headers.get("X-Metabase-Session"))
        with open(save_path, "w", encoding="utf-8") as file:
            file.write(token)

    def _make_request(
        self,
        method: str,
        url: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> Response:
        """Perform an HTTP request, catching and re-raising any exceptions

        Args:
            method (str): GET or POST or DELETE or PUT
            url (str): URL endpoint
            params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Raises:
            RequestFailure: Request failed

        Returns:
            Response: Response from the API
        """
        log_line_pre = f"{method=}, {url=}, {params=}"
        try:
            self._logger.debug(log_line_pre)
            return self._session.request(
                method=method, url=url, params=params, json=json
            )
        except RequestException as error_raised:
            self._logger.error(str(error_raised))
            raise RequestFailure("Request failed") from error_raised

    def generic_request(
        self,
        http_method: str,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """Method for dispatching HTTP requests

        Args:
            http_method (str): GET or POST or PUT or DELETE
            endpoint (str): URL endpoint
            params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Raises:
            InvalidDataReceived: Unable to decode response from API
            AuthenticationFailure: Auth failure received from API
            RequestFailure: Other failure during request

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        log_line_post = "success=%s, status_code=%s, message=%s"
        response = self._make_request(
            method=http_method,
            url=self.metabase_url + endpoint,
            params=params,
            json=json,
        )

        # If status_code in 200-299 range, return Result, else raise exception
        is_success = 299 >= response.status_code >= 200
        if is_success:
            self._logger.debug(
                log_line_post, is_success, response.status_code, response.reason
            )
            try:
                return_ = response.json()
                if isinstance(return_, (list, dict)):
                    return return_
            except JSONDecodeError as error_raised:
                raise InvalidDataReceived from error_raised
        elif response.status_code == 401:
            self._logger.error(
                log_line_post, False, response.status_code, response.text
            )
            raise AuthenticationFailure(f"{response.status_code} - {response.reason}")

        error_line = f"{response.status_code} - {response.reason}"
        self._logger.error(log_line_post)
        raise RequestFailure(error_line)

    def get(
        self, endpoint: str, params: Optional[dict[str, Any]] = None
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP GET request

        Args:
            endpoint (str): URL endpoint
            ep_params (dict, optional): Endpoint parameters

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(http_method="GET", endpoint=endpoint, params=params)

    def post(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP POST request

        Args:
            endpoint (str): URL endpoint
            params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(
            http_method="POST", endpoint=endpoint, params=params, json=json
        )

    def delete(
        self, endpoint: str, params: Optional[dict[str, Any]] = None
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP DELETE request

        Args:
            endpoint (str): URL endpoint
            params (dict, optional): Endpoint parameters

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(
            http_method="DELETE", endpoint=endpoint, params=params
        )

    def put(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP PUT request

        Args:
            endpoint (str): URL endpoint
            ep_params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(
            http_method="PUT", endpoint=endpoint, params=params, json=json
        )



"""
    Generic classes for Metabase
"""
from __future__ import annotations  # Included for support of |

from typing import Any, ClassVar, Optional, TypeVar

from pydantic import BaseModel

from metabase_tools.exceptions import EmptyDataReceived, InvalidParameters
from metabase_tools.metabase import MetabaseApi


class MetabaseGenericObject(BaseModel, extra="forbid"):
    """Base class for all Metabase objects. Provides generic fields and methods."""

    BASE_EP: ClassVar[str]

    id: int | str
    name: str

    @classmethod
    def _request_list(
        cls,
        http_method: str,
        adapter: MetabaseApi,
        endpoint: str,
        source: list[int] | list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Sends requests to API based on a list of objects

        Args:
            http_method (str): GET or POST or PUT or DELETE
            adapter (MetabaseApi): Connection to Metabase API
            endpoint (str): Endpoint to use for request
            source (list[int] | list[dict]): List of targets or payloads

        Raises:
            InvalidParameters: Item in source is not an int or dict
            EmptyDataReceived: No data returned

        Returns:
            list[dict]: Aggregated results of all API calls
        """
        results = []

        for item in source:
            if isinstance(item, int):
                response = adapter.generic_request(
                    http_method=http_method, endpoint=endpoint.format(id=item)
                )
            elif isinstance(item, dict):
                item_ep = endpoint.format(**item)
                if http_method == "PUT":
                    response = adapter.put(
                        endpoint=item_ep,
                        json=item,
                    )
                else:
                    response = adapter.generic_request(
                        http_method=http_method,
                        endpoint=item_ep,
                        json=item,
                    )
            else:
                raise InvalidParameters
            if isinstance(response, dict):
                results.append(response)
            elif isinstance(response, list):
                results.extend(response)
        if len(results) > 0:
            return results
        raise EmptyDataReceived("No data returned")


class GenericWithoutArchive(MetabaseGenericObject):
    """Generic class for objects without support for archive"""

    @classmethod
    def get(
        cls: type[GWA], adapter: MetabaseApi, targets: Optional[list[int]] = None
    ) -> list[GWA]:
        """Generic method for returning an object or list of objects

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int], optional): IDs of the objects being requested

        Raises:
            InvalidParameters: Targets are not None or list[int]
            EmptyDataReceived: No data is received from the API

        Returns:
            list[Self]: List of objects of the relevant type
        """
        if isinstance(targets, list) and all(isinstance(t, int) for t in targets):
            results = cls._request_list(
                http_method="GET",
                adapter=adapter,
                endpoint=cls.BASE_EP + "/{id}",
                source=targets,
            )
            return [cls(**result) for result in results]

        if targets is None:
            # If no targets are provided, all objects of that type should be returned
            response = adapter.get(endpoint=cls.BASE_EP)
            if isinstance(response, list):  # Validate data was returned
                # Unpack data into instances of the class and return
                return [
                    cls(**record) for record in response if isinstance(record, dict)
                ]
        else:
            # If something other than None, int or list[int], raise error
            raise InvalidParameters("Invalid target(s)")
        # If response.data was empty, raise error
        raise EmptyDataReceived("No data returned")

    @classmethod
    def search(
        cls: type[GWA],
        adapter: MetabaseApi,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[GWA]] = None,
    ) -> list[GWA]:
        """Method to search a list of objects meeting a list of parameters

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[Self], optional): Provide to search against an existing \
                list, by default pulls from API

        Returns:
            list[Self]: List of objects of the relevant type
        """
        objs = search_list or cls.get(adapter=adapter)
        results = []
        for param in search_params:
            for obj in objs:
                obj_dict = obj.dict()
                for key, value in param.items():
                    if key in obj_dict and value == obj_dict[key]:
                        results.append(obj)
        return results

    @classmethod
    def create(
        cls: type[GWA], adapter: MetabaseApi, payloads: list[dict[str, Any]]
    ) -> list[GWA]:
        """Generic method for creating a list of objects

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            payloads (list[dict]): List of json payloads

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            list[Self]: List of objects of the relevant type
        """
        if isinstance(payloads, list) and all(isinstance(t, dict) for t in payloads):
            # If a list of targets is provided, return a list of objects
            results = cls._request_list(
                http_method="POST",
                adapter=adapter,
                endpoint=cls.BASE_EP,
                source=payloads,
            )
            return [cls(**result) for result in results]
        # If something other than dict or list[dict], raise error
        raise InvalidParameters("Invalid target(s)")

    @classmethod
    def update(
        cls: type[GWA], adapter: MetabaseApi, payloads: list[dict[str, Any]]
    ) -> list[GWA]:
        """Generic method for updating a list of objects

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            payloads (list[dict]): List of json payloads

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            list[Self]: List of objects of the relevant type
        """
        if isinstance(payloads, list) and all(isinstance(t, dict) for t in payloads):
            # If a list of targets is provided, return a list of objects
            results = cls._request_list(
                http_method="PUT",
                adapter=adapter,
                endpoint=cls.BASE_EP + "/{id}",
                source=payloads,
            )
            return [cls(**result) for result in results]
        # If something other than dict or list[dict], raise error
        raise InvalidParameters("Invalid target(s)")

    @classmethod
    def delete(cls, adapter: MetabaseApi, targets: list[int]) -> dict[int, Any]:
        """Method to delete a list of objects

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of objects to delete

        Raises:
            InvalidParameters: Targets is not a list of ints

        Returns:
            dict: _description_
        """
        if isinstance(targets, list) and all(isinstance(t, int) for t in targets):
            results = cls._request_list(
                http_method="DELETE",
                adapter=adapter,
                endpoint=cls.BASE_EP + "/{id}",
                source=targets,
            )
            return {target: result for result, target in zip(results, targets)}
        raise InvalidParameters("Invalid set of targets")


class GenericWithArchive(GenericWithoutArchive):
    """Generic class for objects with archive support in the API"""

    @classmethod
    def archive(
        cls: type[GA],
        adapter: MetabaseApi,
        targets: list[int],
        unarchive: bool = False,
    ) -> list[GA]:
        """Generic method for archiving a list of objects

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            targets (list[int]): List of objects to archive
            unarchive (bool): Whether object should be unarchived instead of archived

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            list[Self]: List of objects of the relevant type
        """
        if isinstance(targets, list) and all(isinstance(t, int) for t in targets):
            results = cls._request_list(
                http_method="PUT",
                adapter=adapter,
                endpoint=cls.BASE_EP + "/{id}",
                source=[
                    {"id": target, "archived": not unarchive} for target in targets
                ],
            )
            return [cls(**result) for result in results]
        raise InvalidParameters("Invalid set of targets")


MGO = TypeVar("MGO", bound="MetabaseGenericObject")
GWA = TypeVar("GWA", bound="GenericWithoutArchive")
GA = TypeVar("GA", bound="GenericWithArchive")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['metabase_tools', 'metabase_tools.models']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.1,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'types-requests>=2.28.9,<3.0.0']

setup_kwargs = {
    'name': 'metabase-tools',
    'version': '0.6.0',
    'description': 'Unofficial API wrapper for Metabase plus additional helper tools',
    'long_description': None,
    'author': 'Josh Odell',
    'author_email': 'j01101111sh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
